# Nov-27-2023
# magnitude.py

import cv2 as cv
import numpy as np

from pnt3.src import cfg_pnt3


def get_magn_roi(any_image: np.uint8):

    height = any_image.shape[0]
    width = any_image.shape[1]

    array_dft = \
        np.zeros((cfg_pnt3.size_dft, cfg_pnt3.size_dft), dtype=np.float32)

    array_dft[0:height:, 0:width:] = any_image[0::, 0::]

    dft = cv.dft(array_dft, flags=cv.DFT_COMPLEX_OUTPUT)

    dft_shift = np.fft.fftshift(dft)

    re_shift = dft_shift[:, :, 0]
    im_shift = dft_shift[:, :, 1]

    re_roi = np.zeros((cfg_pnt3.size_roi, cfg_pnt3.size_roi), dtype=np.float32)
    im_roi = np.zeros((cfg_pnt3.size_roi, cfg_pnt3.size_roi), dtype=np.float32)
    magn_roi_norm = np.zeros((cfg_pnt3.size_roi, cfg_pnt3.size_roi), dtype=np.float32)

    y0 = (cfg_pnt3.size_dft - cfg_pnt3.size_roi) // 2
    x0 = y0

    re_roi[0:cfg_pnt3.size_roi:, 0:cfg_pnt3.size_roi:] = \
        re_shift[y0:y0+cfg_pnt3.size_roi:, x0:x0+cfg_pnt3.size_roi:]
    im_roi[0:cfg_pnt3.size_roi:, 0:cfg_pnt3.size_roi:] = \
        im_shift[y0:y0+cfg_pnt3.size_roi:, x0:x0+cfg_pnt3.size_roi:]

    magn_roi = cv.magnitude(re_roi, im_roi)

    cv.normalize(magn_roi, magn_roi_norm, 0, 255, cv.NORM_MINMAX)

    return magn_roi_norm


def get_magn_half(magn_roi: np.ndarray):

    size_roi_half_1 = cfg_pnt3.size_roi_half + 1

    shape_magn_half = (size_roi_half_1, cfg_pnt3.size_roi)

    magn_half = np.empty(shape_magn_half, dtype=np.float32)

    magn_half[0:size_roi_half_1, 0:cfg_pnt3.size_roi] = \
        magn_roi[0:size_roi_half_1:, 0::]

    return magn_half



# Nov-27-2023
# draw_canonical.py

import cv2 as cv
from pathlib import Path

from pnt3.src import cfg_pnt3


def draw_canonical(image, templ):

    canon_image_name = 'canon_' + cfg_pnt3.image_name
    canon_templ_name = 'canon_' + cfg_pnt3.templ_name

    path_image_canon = str(Path.cwd() / cfg_pnt3.dir_debug / canon_image_name)
    path_templ_canon = str(Path.cwd() / cfg_pnt3.dir_debug / canon_templ_name)

    cv.imwrite(path_image_canon, image)
    cv.imwrite(path_templ_canon, templ)



# Nov-27-2023
# draw_image_warp.py

import cv2 as cv
import numpy as np
from pathlib import Path

from pnt3.src import cfg_pnt3


def draw_image_warp():

    path_in = str(Path.cwd() / cfg_pnt3.dir_debug / 'image.png')
    # path_in = str(Path.cwd() / cfg_pnt3.dir_debug / 'image_peaks.png')
    image_in = cv.imread(path_in, cv.IMREAD_UNCHANGED)

    y_image_blue = cfg_pnt3.y_image_blue
    x_image_blue = cfg_pnt3.x_image_blue
    y_image_red = cfg_pnt3.y_image_red
    x_image_red = cfg_pnt3.x_image_red

    y_templ_blue = cfg_pnt3.y_templ_blue
    x_templ_blue = cfg_pnt3.x_templ_blue
    y_templ_red = cfg_pnt3.y_templ_red
    x_templ_red = cfg_pnt3.x_templ_red

    pts_image = np.float32([[x_image_blue, y_image_blue],
                            [x_image_red, y_image_red],
                            [cfg_pnt3.X0, cfg_pnt3.Y0]])

    pts_templ = np.float32([[x_templ_blue, y_templ_blue],
                            [x_templ_red, y_templ_red],
                            [cfg_pnt3.X0, cfg_pnt3.Y0]])

    mat_affine = cv.getAffineTransform(pts_image, pts_templ)

    image_out = cv.warpAffine(image_in, mat_affine, cfg_pnt3.dsize_roi)

    path_out = str(Path.cwd() / cfg_pnt3.dir_debug / 'image_warp.png')
    cv.imwrite(path_out, image_out)



# Nov-27-2023
# draw_magn.py

import sys
import os
import cv2 as cv
import numpy as np

from pnt3.src import cfg_pnt3


def draw_magn(tag, any_image_magn):

    # Save Magnitude in file as grayscale.
    # ---------------------------------------------------------
    magnitude_gray_path = cfg_pnt3.dir_debug + '/' + tag + '_gray.png'
    result = cv.imwrite(magnitude_gray_path, any_image_magn)

    if result is False:
        print('ERROR: draw_magn()')
        sys.exit(1)
    # ---------------------------------------------------------

    # Convert Magnitude to ColorMap.
    # ---------------------------------------------------------
    magnitude_color = np.uint8(0)

    if os.path.exists(magnitude_gray_path):

        magnitude_gray = cv.imread(magnitude_gray_path, cv.IMREAD_GRAYSCALE)

        os.remove(magnitude_gray_path)

        if cfg_pnt3.colormap == 'HOT':
            magnitude_color = cv.applyColorMap(magnitude_gray, cv.COLORMAP_HOT)

        if cfg_pnt3.colormap == 'HSV':
            magnitude_color = cv.applyColorMap(magnitude_gray, cv.COLORMAP_HSV)

        if cfg_pnt3.colormap == 'JET':
            magnitude_color = cv.applyColorMap(magnitude_gray, cv.COLORMAP_JET)

        if cfg_pnt3.colormap == 'TURBO':
            magnitude_color = cv.applyColorMap(magnitude_gray, cv.COLORMAP_TURBO)
    else:
        print("ERROR in draw_magn(): The file does not exist")
        sys.exit(1)
    # ---------------------------------------------------------

    magnitude_color_path = cfg_pnt3.dir_debug + '/' + tag + '.png'
    cv.imwrite(magnitude_color_path, magnitude_color)



# Nov-27-2023
# draw_utils.py

import cv2 as cv

from pnt3.src import cfg_pnt3


def draw_peak(any_image, x, y, color, size, thickness):
    start_point = (x - size, y - size)
    end_point = (x + size, y + size)
    cv.rectangle(
        any_image, start_point, end_point,
        color, thickness)


def draw_text(any_image, x, y, text_color, text):

    font = cv.FONT_HERSHEY_SIMPLEX
    font_scale = 0.6
    font_color = text_color
    thickness = 1
    cv.putText(
        any_image, text, (x, y),
        font, font_scale, font_color,
        thickness, cv.LINE_AA)


def draw_peak_color(magnitude, x, y, param):

    thickness = -1
    color = cfg_pnt3.black

    if param < 8:
        size = 9 - param

        if param == 0:
            color = cfg_pnt3.nickel
        if param == 1:
            color = cfg_pnt3.cayenne
        if param == 2:
            color = cfg_pnt3.moss
        if param == 3:
            color = cfg_pnt3.ocean
        if param == 4:
            color = cfg_pnt3.magenta
        if param == 5:
            color = cfg_pnt3.blueberry
        if param == 6:
            color = cfg_pnt3.cayenne
        if param == 7:
            color = cfg_pnt3.nickel

        start_point = (x - size, y - size)
        end_point = (x + size, y + size)
        cv.rectangle(
            magnitude, start_point, end_point,
            color, thickness)
    else:
        radius = 3
        cv.circle(
            magnitude, (x, y), radius,
            color, thickness)



# Nov-27-2023
# get_angle.py

"""
Система координат изображения в OpenCV (opencv-coordinates)

The coordinate origin is assumed to be the top-left corner. Rotation angle in degrees.
Positive values mean counter-clockwise rotation (против часовой стрелки).


0----------------> X
|
|
|
|
|
Y


Система координат централизованная (center-coordinates)

         Y
         ^
         |
         |
---------0--------> X
         |
         |
         |
         |
"""

import math

from pnt3.src import cfg_pnt3


def get_angle(x_a, y_a, x_b, y_b) -> float:

    """
    (x_a, y_a), (x_b, y_b) - координаты двух локальных максимумов спектра
    (пиков) в системе координат изображения (opencv-coordinates).

    Проводим два вектора из центра изображения в точки нахождения этих пиков.
    Функция вычисляет угол между этими векторами в централизованной системе
    координат.

    Диапазон измерения угла [0 - 180] градусов. Результат всегда >= 0,
    т.е. угол не зависит от порядка вводимых векторов.
    """

    # Координаты пиков в централизованной системе координат.
    _x1 = float(x_a - cfg_pnt3.X0)
    _y1 = float(cfg_pnt3.Y0 - y_a)

    _x2 = float(x_b - cfg_pnt3.X0)
    _y2 = float(cfg_pnt3.Y0 - y_b)

    top = (_x1 * _x2) + (_y1 * _y2)
    bottom = math.sqrt((_x1 * _x1) + (_y1 * _y1)) * math.sqrt((_x2 * _x2) + (_y2 * _y2))

    if bottom > 0.0:
        cos_angle = top / bottom

        if cos_angle > 1.0:
            cos_angle = 1.0
        if cos_angle < -1.0:
            cos_angle = -1.0

        angle_rad = math.acos(cos_angle)
        angle_deg = math.degrees(angle_rad)
    else:
        angle_deg = 0.0

    return angle_deg


def get_angle_axe_x(x, y) -> float:

    """
    x, y - координаты локального максимума спектра (пика) в
    системе координат изображения (opencv-coordinates).

    Проводим вектор из центра изображения в точку нахождения пика.
    Функция вычисляет угол между этим вектором и осью X в централизованной
    системе координат.
    Диапазон [0 - 360] градусов. Результат всегда >= 0.
    """

    # Координаты пиков в централизованной системе координат.
    _x1 = float(x - cfg_pnt3.X0)
    _y1 = float(cfg_pnt3.Y0 - y)

    _x2 = 1.0
    _y2 = 0.0

    top = (_x1 * _x2) + (_y1 * _y2)
    bottom = math.sqrt((_x1 * _x1) + (_y1 * _y1)) * math.sqrt((_x2 * _x2) + (_y2 * _y2))

    if bottom > 0.0:
        cos_angle = top / bottom

        if cos_angle > 1.0:
            cos_angle = 1.0
        if cos_angle < -1.0:
            cos_angle = -1.0

        angle_rad = math.acos(cos_angle)
        angle_deg = math.degrees(angle_rad)

        if _y1 < 0.0:
            angle_deg = 360.0 - angle_deg
    else:
        angle_deg = 0.0

    return angle_deg




# Nov-27-2023
# draw_result.py

import cv2 as cv

from pnt3.src import cfg_pnt3


def draw_result(tag):

    path_in = cfg_pnt3.dir_debug + '/' + tag + '_peaks.png'
    magn_peaks = cv.imread(path_in, cv.IMREAD_UNCHANGED)

    x1: int = 0
    y1: int = 0
    x2: int = 0
    y2: int = 0
    x3: int = cfg_pnt3.X0
    y3: int = cfg_pnt3.Y0

    if tag == 'image':
        x1 = cfg_pnt3.x_image_red
        y1 = cfg_pnt3.y_image_red
        x2 = cfg_pnt3.x_image_blue
        y2 = cfg_pnt3.y_image_blue

    if tag == 'templ':
        x1 = cfg_pnt3.x_templ_red
        y1 = cfg_pnt3.y_templ_red
        x2 = cfg_pnt3.x_templ_blue
        y2 = cfg_pnt3.y_templ_blue

    thickness = 1

    """
    cv.line(magn_peaks,
            (x1, y1),
            (x2, y2),
            cfg_pnt3.clover, thickness)  # green

    cv.line(magn_peaks,
            (x1, y1),
            (x3, y3),
            cfg_pnt3.maraschino, thickness)  # red

    cv.line(magn_peaks,
            (x2, y2),
            (x3, y3),
            cfg_pnt3.aqua, thickness)  # blue
    """

    cv.line(magn_peaks,
            (x1, y1),
            (x2, y2),
            cfg_pnt3.black, thickness)

    cv.line(magn_peaks,
            (x1, y1),
            (x3, y3),
            cfg_pnt3.black, thickness)

    cv.line(magn_peaks,
            (x2, y2),
            (x3, y3),
            cfg_pnt3.black, thickness)

    path_out = cfg_pnt3.dir_debug + '/' + tag + '_result.png'
    cv.imwrite(path_out, magn_peaks)



# Nov-27-2023
# utils.py

from pathlib import Path


def init_directory(dir_name):

    path_dir = Path.cwd() / dir_name

    if path_dir.is_dir():
        for child in path_dir.glob('*'):
            if child.is_file():
                child.unlink()
    else:
        create_directory(dir_name)


def create_directory(dir_name):

    path_dir = Path.cwd() / dir_name

    if not path_dir.is_dir():
        path_dir.mkdir()


def remove_directory(dir_name):

    path_dir = Path.cwd() / dir_name

    if path_dir.is_dir():
        for child in path_dir.glob('*'):
            if child.is_file():
                child.unlink()
        path_dir.rmdir()



# Nov-27-2023
# create_joint_table.py

import numpy as np
from numpy import ndarray

from pnt3.src import cfg_pnt3
from pnt3.src.get_angle import get_angle


def create_joint_table(
        image_table, n_image_table_rows,
        templ_table, n_templ_table_rows) -> (ndarray, int):

    n_joint_table_rows = n_image_table_rows * n_templ_table_rows

    joint_table = np.zeros((n_joint_table_rows, 8), dtype=np.int32)

    n_rows: int = 0
    for n_image in range(n_image_table_rows):

        y_image_blue = image_table[n_image, 0]
        x_image_blue = image_table[n_image, 1]
        y_image_red = image_table[n_image, 2]
        x_image_red = image_table[n_image, 3]

        angle_image = \
            get_angle(x_image_blue, y_image_blue, x_image_red, y_image_red)

        if angle_image == float(0):
            continue

        for n_templ in range(n_templ_table_rows):

            y_templ_blue = templ_table[n_templ, 0]
            x_templ_blue = templ_table[n_templ, 1]
            y_templ_red = templ_table[n_templ, 2]
            x_templ_red = templ_table[n_templ, 3]

            angle_templ = \
                get_angle(x_templ_blue, y_templ_blue, x_templ_red, y_templ_red)
            
            if angle_templ == float(0):
                continue

            if angle_image > angle_templ:
                angle_ratio = angle_image / angle_templ
            else:
                angle_ratio = angle_templ / angle_image

            if angle_ratio > cfg_pnt3.angle_ratio_threshold:
                continue

            joint_table[n_rows, 0] = y_image_blue
            joint_table[n_rows, 1] = x_image_blue
            joint_table[n_rows, 2] = y_image_red
            joint_table[n_rows, 3] = x_image_red
            joint_table[n_rows, 4] = y_templ_blue
            joint_table[n_rows, 5] = x_templ_blue
            joint_table[n_rows, 6] = y_templ_red
            joint_table[n_rows, 7] = x_templ_red

            n_rows += 1

    return joint_table, n_rows



# Nov-27-2023
# calc_similarity.py

import sys
import cv2 as cv
import numpy as np

from pnt3.src import cfg_pnt3
from pnt3.src.magnitude import get_magn_roi, get_magn_half
from pnt3.src.get_peaks import get_peaks
from pnt3.src.create_table import create_table
from pnt3.src.create_joint_table import create_joint_table
from pnt3.src.draw_magn import draw_magn
from pnt3.src.draw_peaks import draw_peaks
from pnt3.src.draw_result import draw_result
from pnt3.src.draw_image_warp import draw_image_warp
from pnt3.src.draw_similarity_map import draw_similarity_map
from pnt3.src.utils_list import save_list_of_peaks_txt


def calc_similarity(image: np.uint8, templ: np.uint8) -> float:

    image_magn_roi = get_magn_roi(image)
    image_magn_half = get_magn_half(image_magn_roi)
    image_list_of_peaks = get_peaks(image_magn_half)
    image_table, n_image_table_rows = create_table(image_list_of_peaks)

    if n_image_table_rows == 0:
        print(f'\ncalc_similarity ERROR: ' + cfg_pnt3.path_image + ' - unacceptable shape')
        sys.exit(1)

    if cfg_pnt3.debug_mode:
        draw_magn('image', image_magn_roi)
        draw_peaks('image', image_list_of_peaks)
        save_list_of_peaks_txt('image_list_of_peaks', image_list_of_peaks)

    templ_magn_roi = get_magn_roi(templ)
    templ_magn_half = get_magn_half(templ_magn_roi)
    templ_list_of_peaks = get_peaks(templ_magn_half)
    templ_table, n_templ_table_rows = create_table(templ_list_of_peaks)

    if n_templ_table_rows == 0:
        print(f'\ncalc_similarity ERROR: ' + cfg_pnt3.path_templ + ' - unacceptable shape')
        sys.exit(1)

    if cfg_pnt3.debug_mode:
        draw_magn('templ', templ_magn_roi)
        draw_peaks('templ', templ_list_of_peaks)
        save_list_of_peaks_txt('templ_list_of_peaks', templ_list_of_peaks)

    joint_table, n_joint_table_rows = \
        create_joint_table(image_table, n_image_table_rows,
                           templ_table, n_templ_table_rows)

    if n_joint_table_rows == 0:
        print(f"\nERROR: " + cfg_pnt3.path_image + "  vs.  " + cfg_pnt3.path_templ + " - we can't compare")
        sys.exit(1)

    similarity = calc_similarity_cont(
                        image_magn_roi, templ_magn_roi,
                        joint_table, n_joint_table_rows)

    return similarity


def calc_similarity_cont(
        image_magn_roi, templ_magn_roi,
        joint_table, joint_table_rows) -> float:

    similarity_map_current: np.float32 = np.float32(0)
    similarity_map_result: np.float32 = np.float32(0)

    similarity_result: float = -1

    for n in range(joint_table_rows):

        y_image_blue = joint_table[n, 0]
        x_image_blue = joint_table[n, 1]
        y_image_red = joint_table[n, 2]
        x_image_red = joint_table[n, 3]

        y_templ_blue = joint_table[n, 4]
        x_templ_blue = joint_table[n, 5]
        y_templ_red = joint_table[n, 6]
        x_templ_red = joint_table[n, 7]

        pts_image = np.float32([[x_image_blue, y_image_blue],
                                [x_image_red, y_image_red],
                                [cfg_pnt3.X0, cfg_pnt3.Y0]])

        pts_templ = np.float32([[x_templ_blue, y_templ_blue],
                                [x_templ_red, y_templ_red],
                                [cfg_pnt3.X0, cfg_pnt3.Y0]])

        mat_affine = cv.getAffineTransform(pts_image, pts_templ)
        dsize = (cfg_pnt3.size_roi, cfg_pnt3.size_roi)

        image_magn_roi_warp = cv.warpAffine(image_magn_roi, mat_affine, dsize)

        if cfg_pnt3.debug_mode:
            similarity_current, similarity_map_current \
                                    = calc_similarity_cont2_map(
                                            np.float32(image_magn_roi_warp),
                                            np.float32(templ_magn_roi))
        else:
            similarity_current = calc_similarity_cont2(
                                            np.float32(image_magn_roi_warp),
                                            np.float32(templ_magn_roi))

        if similarity_current > similarity_result:
            similarity_result = similarity_current

            if cfg_pnt3.debug_mode:
                similarity_map_result = similarity_map_current.copy()

                cfg_pnt3.y_image_blue = y_image_blue
                cfg_pnt3.x_image_blue = x_image_blue
                cfg_pnt3.y_image_red = y_image_red
                cfg_pnt3.x_image_red = x_image_red

                cfg_pnt3.y_templ_blue = y_templ_blue
                cfg_pnt3.x_templ_blue = x_templ_blue
                cfg_pnt3.y_templ_red = y_templ_red
                cfg_pnt3.x_templ_red = x_templ_red

    if cfg_pnt3.debug_mode:
        draw_result('image')
        draw_result('templ')
        draw_image_warp()
        draw_similarity_map(similarity_map_result)

    return similarity_result


def calc_similarity_cont2_map(
        magnitude_1: np.float32,
        magnitude_2: np.float32) -> (float, np.float32):

    arr_zero = np.zeros(cfg_pnt3.dsize_roi, dtype=np.float32)
    arr_1 = np.maximum(magnitude_1, arr_zero)
    arr_2 = np.maximum(magnitude_2, arr_zero)

    arr_min = np.minimum(arr_1, arr_2)
    arr_max = np.maximum(arr_1, arr_2)

    similarity_map = np.subtract(arr_max, arr_min, where=arr_min > 0)
    nonzero_count = np.count_nonzero(similarity_map)

    if nonzero_count == 0:
        similarity = 1.0
    else:
        average = np.sum(similarity_map) / nonzero_count
        similarity = 1.0 / (1.0 + cfg_pnt3.param_similarity * average)

    return similarity, similarity_map


def calc_similarity_cont2(
        magnitude_1: np.float32,
        magnitude_2: np.float32) -> float:

    arr_zero = np.zeros(cfg_pnt3.dsize_roi, dtype=np.float32)
    arr_1 = np.maximum(magnitude_1, arr_zero)
    arr_2 = np.maximum(magnitude_2, arr_zero)

    arr_min = np.minimum(arr_1, arr_2)
    arr_max = np.maximum(arr_1, arr_2)

    similarity_map = np.subtract(arr_max, arr_min, where=arr_min > 0)
    nonzero_count = np.count_nonzero(similarity_map)

    if nonzero_count == 0:
        similarity = 1.0
    else:
        average = np.sum(similarity_map) / nonzero_count
        similarity = 1.0 / (1.0 + cfg_pnt3.param_similarity * average)

    return similarity



# Nov-27-2023
# get_similarity.py

from pnt3.src import cfg_pnt3
from pnt3.src.canonical import convert_to_canonical
from pnt3.src.calc_similarity import calc_similarity
from pnt3.src.draw_canonical import draw_canonical


def get_similarity(path_1: str, path_2: str) -> float:

    image = convert_to_canonical('image', path_1)
    templ = convert_to_canonical('templ', path_2)

    if cfg_pnt3.debug_mode:
        draw_canonical(image, templ)

    similarity = calc_similarity(image, templ)

    return similarity



# Nov-27-2023
# draw_similarity_map.py

import sys
import os
import cv2 as cv

from pnt3.src import cfg_pnt3


def draw_similarity_map(similarity_map):

    # Save Magnitude in file as grayscale.
    # ---------------------------------------------------------
    similarity_map_gray_path = cfg_pnt3.dir_debug + '/' + '_similarity_map_gray.png'
    result = cv.imwrite(similarity_map_gray_path, similarity_map)

    if result is False:
        print('ERROR: draw_similarity_map()')
        sys.exit(1)
    # ---------------------------------------------------------

    # Convert similarity_map to ColorMap.
    # ---------------------------------------------------------
    if os.path.exists(similarity_map_gray_path):

        similarity_map_gray = cv.imread(similarity_map_gray_path, cv.IMREAD_GRAYSCALE)

        os.remove(similarity_map_gray_path)

        similarity_map_color = cv.applyColorMap(similarity_map_gray, cv.COLORMAP_TURBO)
    else:
        print("ERROR in draw_similarity_map(): The file does not exist")
        sys.exit(1)
    # ---------------------------------------------------------

    similarity_map_color_path = cfg_pnt3.dir_debug + '/' + '_similarity_map.png'
    cv.imwrite(similarity_map_color_path, similarity_map_color)



# Nov-27-2023
# draw_peaks.py

import cv2 as cv

from pnt3.src import cfg_pnt3
from pnt3.src.draw_utils import draw_peak, draw_text


def draw_peaks(tag, list_of_peaks):

    path_in = cfg_pnt3.dir_debug + '/' + tag + '.png'
    magn = cv.imread(path_in, cv.IMREAD_UNCHANGED)

    # Draw peaks
    # -----------------------------------------
    n = 0
    for item in list_of_peaks:
        y = int(item[0])
        x = int(item[1])
        draw_peak(magn, x, y, cfg_pnt3.black, 2, -1)
        if n > 0:
            draw_text(magn, x + 5, y + 0, cfg_pnt3.black, str(n))
        n += 1
    # -----------------------------------------

    path_out = cfg_pnt3.dir_debug + '/' + tag + '_peaks.png'
    cv.imwrite(path_out, magn)



# Nov-27-2023
# utils_list.py

from pnt3.src import cfg_pnt3


def save_list_of_peaks_txt(tag, list_of_peaks):

    data_path = cfg_pnt3.dir_debug + '/' + tag + '.txt'

    with open(data_path, 'w') as fp:
        for item in list_of_peaks:
            fp.write(f'{item}\n')


def read_list_of_peaks_txt(tag):

    list_of_peaks = []

    data_path = cfg_pnt3.dir_debug + '/' + tag + '.txt'

    with open(data_path, 'r') as fp:
        for line in fp:

            # remove linebreak from a current item
            item = line[:-1]

            list_of_peaks.append(item)

    return list_of_peaks


def print_list(tag, any_list):
    print(f'\n{tag}')
    n = 0
    for item in any_list:
        print(f'{n}:\t {item}')
        n += 1



# Nov-27-2023
# create_table.py

import sys
import math
import numpy as np
from numpy import ndarray
from itertools import combinations

from pnt3.src import cfg_pnt3


def create_table(list_of_peaks) -> (ndarray, int):

    first = list_of_peaks[0]
    y_peaks_0 = first[0]
    x_peaks_0 = first[1]

    if y_peaks_0 != cfg_pnt3.Y0 or x_peaks_0 != cfg_pnt3.X0:
        print(f'\nSomething is going wrong')
        sys.exit(1)

    list_of_peaks_half = []
    for item in list_of_peaks:
        if item[0] > cfg_pnt3.size_roi_half:
            continue
        list_of_peaks_half.append(item)

    list_of_peaks_1 = list_of_peaks_half[1:]

    arr_out = np.array(list(combinations(list_of_peaks_1, 2)))
    n_combinations = arr_out.shape[0]

    table_rows = n_combinations
    table_cols = 4

    table = np.zeros((table_rows, table_cols), dtype=np.int32)

    n_rows: int = 0
    for n in range(table_rows):

        item_1 = arr_out[n, 0]
        item_2 = arr_out[n, 1]

        y1 = int(item_1[0])
        x1 = int(item_1[1])

        y2 = int(item_2[0])
        x2 = int(item_2[1])

        """
        Если двигаться против часовой стрелки, то точка "blue" 
        расположена после точки "red".
        """
        x_blue, y_blue, x_red, y_red = set_order(x1, y1, x2, y2)

        if x_blue == -1:
            continue

        table[n_rows, 0] = y_blue
        table[n_rows, 1] = x_blue
        table[n_rows, 2] = y_red
        table[n_rows, 3] = x_red

        n_rows += 1

    return table, n_rows


def set_order(x1: int, y1: int, x2: int, y2: int) -> \
        [int, int, int, int]:

    a1 = get_angle_axe_x(x1, y1)
    a2 = get_angle_axe_x(x2, y2)

    da = abs(a1 - a2)

    x_blue = -1
    y_blue = -1
    x_red = -1
    y_red = -1

    if (da >= cfg_pnt3.angle_min) and (da <= cfg_pnt3.angle_max):
        if a2 > a1:
            x_blue = x2
            y_blue = y2
            x_red = x1
            y_red = y1
        else:
            x_blue = x1
            y_blue = y1
            x_red = x2
            y_red = y2

    return [x_blue, y_blue, x_red, y_red]


def get_angle_axe_x(x, y) -> float:

    _x1 = float(x - cfg_pnt3.X0)
    _y1 = float(cfg_pnt3.Y0 - y)

    _x2 = 1.0
    _y2 = 0.0

    top = (_x1 * _x2) + (_y1 * _y2)
    bottom = math.sqrt((_x1 * _x1) + (_y1 * _y1)) * math.sqrt((_x2 * _x2) + (_y2 * _y2))

    if bottom > 0.0:
        cos_angle = top / bottom

        if cos_angle > 1.0:
            cos_angle = 1.0
        if cos_angle < -1.0:
            cos_angle = -1.0

        angle_rad = math.acos(cos_angle)
        angle_deg = math.degrees(angle_rad)

        if _y1 < 0.0:
            angle_deg = 360.0 - angle_deg
    else:
        angle_deg = 0.0

    return angle_deg



# Nov-27-2023
# canonical.py

import cv2 as cv
import numpy as np

from pnt3.src import cfg_pnt3


def convert_to_canonical(tag: str, path: str) -> np.uint8:
    # ---------------------------------------------------------
    if tag == 'image':
        cfg_pnt3.path_image = path
    if tag == 'templ':
        cfg_pnt3.path_templ = path
    # ---------------------------------------------------------
    image_gray = cv.imread(path, cv.IMREAD_GRAYSCALE)

    image_canonical = to_canonical(image_gray)
    # ---------------------------------------------------------
    return image_canonical


def to_canonical(image_gray) -> np.uint8:
    # ---------------------------------------------------------
    width = image_gray.shape[1]
    height = image_gray.shape[0]

    max_original_size = max(height, width)
    if max_original_size > cfg_pnt3.canonical_size:
        scale = float(cfg_pnt3.canonical_size) / float(max_original_size)
        width: int = int(width * scale)
        height: int = int(height * scale)
        dim = (width, height)
        image_scale = cv.resize(image_gray, dim, cv.INTER_LANCZOS4)
        image_invert = np.invert(image_scale)
    else:
        image_invert = np.invert(image_gray)
    # ---------------------------------------------------------
    image_canonical = np.zeros((height, width), dtype=np.uint8)
    cv.normalize(image_invert, image_canonical, 0, 255, cv.NORM_MINMAX)
    # ---------------------------------------------------------
    return np.uint8(image_canonical)
